/* -*- c-file-style: "linux" -*-
 * ssl.c: operations for negotiating SSL rsync connections.
 *
 * Copyright (C) 2003  Casey Marshall <rsdio@metastatic.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "rsync.h"

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#endif
#include <string.h>

#define BUF_SIZE 1024

extern int verbose;
extern int am_daemon;
extern int am_server;

extern char *ssl_cert_path;
extern char *ssl_key_path;
extern char *ssl_key_passwd;
extern char *ssl_ca_path;

static SSL_CTX *ssl_ctx;
static SSL *ssl;
static int tls_read[2] = { -1, -1 };
static int tls_write[2] = { -1, -1 };
static int ssl_running;
static int ssl_pid = -1;

#ifdef HAVE_SIGACTION
static struct sigaction sigact;
#endif

/**
 * A non-interactive callback to be passed to SSL_CTX_set_default_password_cb,
 * which merely copies the value of ssl_key_passwd into buf. This is
 * used for when the private key password is supplied via an option.
 */
static int default_password_cb(char *buf, int n, UNUSED(int f), UNUSED(void *u))
{
	if (ssl_key_passwd == NULL || n < (int)strlen(ssl_key_passwd))
		return 0;
	strncpy(buf, ssl_key_passwd, n-1);
	return strlen(ssl_key_passwd);
}

/**
 * If verbose, this method traces the status of the SSL handshake.
 */
static void info_callback(const SSL *ssl, int cb, int val)
{
	char buf[128];
	char *cbs;

	switch (cb) {
	case SSL_CB_LOOP:
		cbs = "SSL_CB_LOOP";
		break;
	case SSL_CB_EXIT:
		cbs = "SSL_CB_EXIT";
		break;
	case SSL_CB_READ:
		cbs = "SSL_CB_READ";
		break;
	case SSL_CB_WRITE:
		cbs = "SSL_CB_WRITE";
		break;
	case SSL_CB_ALERT:
		cbs = "SSL_CB_ALERT";
		break;
	case SSL_CB_READ_ALERT:
		cbs = "SSL_CB_READ_ALERT";
		break;
	case SSL_CB_WRITE_ALERT:
		cbs = "SSL_CB_WRITE_ALERT";
		break;
	case SSL_CB_ACCEPT_LOOP:
		cbs = "SSL_CB_ACCEPT_LOOP";
		break;
	case SSL_CB_ACCEPT_EXIT:
		cbs = "SSL_CB_ACCEPT_EXIT";
		break;
	case SSL_CB_CONNECT_LOOP:
		cbs = "SSL_CB_CONNECT_LOOP";
		break;
	case SSL_CB_CONNECT_EXIT:
		cbs = "SSL_CB_CONNECT_EXIT";
		break;
	case SSL_CB_HANDSHAKE_START:
		cbs = "SSL_CB_HANDSHAKE_START";
		break;
	case SSL_CB_HANDSHAKE_DONE:
		cbs = "SSL_CB_HANDSHAKE_DONE";
		break;
	default:
		snprintf(buf, sizeof buf, "??? (%d)", cb);
		cbs = buf;
		break;
	}
	if (verbose > 2) {
		rprintf(FLOG, "SSL: info_callback(%p,%s,%d)\n", ssl, cbs, val);
		if (cb == SSL_CB_HANDSHAKE_DONE) {
			SSL_CIPHER_description(SSL_get_current_cipher((SSL*)ssl),
					       buf, sizeof buf);
			rprintf(FLOG, "SSL: cipher: %s", buf);
		}
	}
}

/**
 * Initializes the SSL context for TLSv1 connections; returns zero on
 * success.
 */
int init_tls(void)
{
	if (ssl_ctx)
		return 0;
	SSL_library_init();
	SSL_load_error_strings();
	ssl_ctx = SSL_CTX_new(TLSv1_method());
	if (!ssl_ctx)
		return 1;
	SSL_CTX_set_info_callback(ssl_ctx, info_callback);

	/* Sets the certificate sent to the other party. */
	if (ssl_cert_path != NULL
	    && SSL_CTX_use_certificate_file(ssl_ctx, ssl_cert_path,
					    SSL_FILETYPE_PEM) != 1)
		return 1;
	/* Set up the simple non-interactive callback if the password
	 * was supplied on the command line. */
	if (ssl_key_passwd != NULL)
		SSL_CTX_set_default_passwd_cb(ssl_ctx, default_password_cb);
	/* Sets the private key that matches the public certificate. */
	if (ssl_key_path != NULL) {
		if (SSL_CTX_use_PrivateKey_file(ssl_ctx, ssl_key_path,
						SSL_FILETYPE_PEM) != 1)
			return 1;
		if (SSL_CTX_check_private_key(ssl_ctx) != 1)
			return 1;
	}
	if (ssl_ca_path != NULL
	    && !SSL_CTX_load_verify_locations(ssl_ctx, ssl_ca_path, NULL))
		return 1;

	return 0;
}

/**
 * Returns the error string for the current SSL error, if any.
 */
char *get_ssl_error(void)
{
	return ERR_error_string(ERR_get_error(), NULL);
}

/**
 * Returns the input file descriptor for the SSL connection.
 */
int get_tls_rfd(void)
{
	return tls_read[0];
}

/**
 * Returns the output file descriptor for the SSL connection.
 */
int get_tls_wfd(void)
{
	return tls_write[1];
}

/**
 * Signal handler that ends the SSL connection.
 */
static RETSIGTYPE tls_sigusr1(int UNUSED(val))
{
	if (ssl) {
		SSL_shutdown(ssl);
		SSL_free(ssl);
		ssl = NULL;
	}
	ssl_running = 0;
}

/**
 * Negotiates the TLS connection, creates a socket pair for communicating
 * with the rsync process, then forks into a new process that will handle
 * the communication.
 *
 * 0 is returned on success.
 */
int start_tls(int f_in, int f_out)
{
	int tls_fd;
	int n = 0, r;
	unsigned char buf1[BUF_SIZE], buf2[BUF_SIZE];
	int avail1 = 0, avail2 = 0, write1 = 0, write2 = 0;
	fd_set rd, wd;

	if (fd_pair(tls_read))
		return 1;
	if (fd_pair(tls_write))
		return 1;

	set_blocking(tls_read[0]);
	set_blocking(tls_read[1]);
	set_blocking(tls_write[0]);
	set_blocking(tls_write[1]);
	set_blocking(f_in);
	set_blocking(f_out);

	ssl_pid = do_fork();
	if (ssl_pid < 0)
		return -1;
	if (ssl_pid != 0) {
		close(tls_write[0]);
		close(tls_read[1]);
		return 0;
	}

	SIGACTION(SIGUSR1, tls_sigusr1);
	ssl = SSL_new(ssl_ctx);
	if (!ssl)
		goto closed;
	if (am_daemon || am_server)
		SSL_set_accept_state(ssl);
	else
		SSL_set_connect_state(ssl);
	SSL_set_rfd(ssl, f_in);
	SSL_set_wfd(ssl, f_out);

	tls_fd = SSL_get_fd(ssl);
	n = tls_write[0];
	n = MAX(tls_read[1], n);
	n = MAX(tls_fd, n) + 1;

	ssl_running = 1;
	while (ssl_running) {
		FD_ZERO(&rd);
		FD_ZERO(&wd);
		FD_SET(tls_write[0], &rd);
		FD_SET(tls_read[1], &wd);
		FD_SET(tls_fd, &rd);
		FD_SET(tls_fd, &wd);

		r = select(n, &rd, &wd, NULL, NULL);

		if (r == -1 && errno == EINTR)
			continue;
		if (FD_ISSET(tls_write[0], &rd)) {
			r = read(tls_write[0], buf1+avail1, BUF_SIZE-avail1);
			if (r >= 0)
				avail1 += r;
			else {
				rprintf(FERROR, "pipe read error: %s\n",
					strerror(errno));
				break;
			}
		}
		if (FD_ISSET(tls_fd, &rd)) {
			r = SSL_read(ssl, buf2+avail2, BUF_SIZE-avail2);
			if (r > 0)
				avail2 += r;
			else {
				switch (SSL_get_error(ssl, r)) {
				case SSL_ERROR_ZERO_RETURN:
					goto closed;
				case SSL_ERROR_WANT_READ:
				case SSL_ERROR_WANT_WRITE:
					break;
				case SSL_ERROR_SYSCALL:
					if (r == 0)
						rprintf(FERROR, "SSL spurious EOF\n");
					else
						rprintf(FERROR, "SSL I/O error: %s\n",
							strerror(errno));
					goto closed;
				case SSL_ERROR_SSL:
					rprintf(FERROR, "SSL: %s\n",
						ERR_error_string(ERR_get_error(), NULL));
					goto closed;
				default:
					rprintf(FERROR, "unexpected ssl error %d\n", r);
					goto closed;
				}
			}
		}
		if (FD_ISSET(tls_read[1], &wd) && write2 < avail2) {
			r = write(tls_read[1], buf2+write2, avail2-write2);
			if (r >= 0)
				write2 += r;
			else {
				rprintf(FERROR, "pipe write error: %s\n",
					strerror(errno));
				break;
			}
		}
		if (FD_ISSET(tls_fd, &wd) && write1 < avail1) {
			r = SSL_write(ssl, buf1+write1, avail1-write1);
			if (r > 0)
				write1 += r;
			else {
				switch (SSL_get_error(ssl, r)) {
				case SSL_ERROR_ZERO_RETURN:
					goto closed;
				case SSL_ERROR_WANT_READ:
				case SSL_ERROR_WANT_WRITE:
					break;
				case SSL_ERROR_SYSCALL:
					if (r == 0)
						rprintf(FERROR, "SSL: spurious EOF\n");
					else
						rprintf(FERROR, "SSL: I/O error: %s\n",
							strerror(errno));
					goto closed;
				case SSL_ERROR_SSL:
					rprintf(FERROR, "SSL: %s\n",
						ERR_error_string(ERR_get_error(), NULL));
					goto closed;
				default:
					rprintf(FERROR, "unexpected ssl error %d\n", r);
					goto closed;
				}
			}
		}
		if (avail1 == write1)
			avail1 = write1 = 0;
		if (avail2 == write2)
			avail2 = write2 = 0;
	}

	/* XXX I'm pretty sure that there is a lot that I am not considering
	   here. Bugs? Yes, probably. */

	/* We're finished. */
    closed:
	close(tls_read[1]);
	close(tls_write[0]);
	exit(0);
}

/**
 * Ends the TLS connection.
 */
void end_tls(void)
{
	if (ssl_pid > 0)
		kill(ssl_pid, SIGUSR1);
}
